use crate::api;
use crate::display::create_display_thread;
use crate::dither::DitheredImage;
use axum::extract::{Path, State};
use axum::http::{header, StatusCode};
use axum::response::{IntoResponse, Response};
use axum::routing::get;
use axum::Router;
use include_dir::{include_dir, Dir, DirEntry};
use minijinja::Environment;
use std::sync::mpsc;
use std::sync::Arc;
use std::thread::JoinHandle;
use tracing::{debug, warn};
use tracing::{error, info, instrument};

#[derive(thiserror::Error, Debug)]
pub enum ApiError {
    #[error("missing image field")]
    MissingImage,
}

#[derive(Clone)]
pub struct AppState {
    pub display_channel: mpsc::Sender<Box<DitheredImage>>,
    pub display_task: Arc<JoinHandle<()>>,
    pub templates: Arc<Environment<'static>>,
}

impl Default for AppState {
    #[must_use]
    fn default() -> Self {
        let (handle, tx) = create_display_thread();
        Self {
            display_channel: tx,
            display_task: Arc::new(handle),
            templates: Arc::new(make_environment().expect("valid")),
        }
    }
}

/// Represents an error that can be returned to the requester by implementing ``IntoResponse``.
/// This wraps ``anyhow::Error`` so that it can accept any error object.
pub struct AppError(anyhow::Error);

// Tell axum how to convert `AppError` into a response.
impl IntoResponse for AppError {
    fn into_response(self) -> Response {
        (
            StatusCode::INTERNAL_SERVER_ERROR,
            format!("Something went wrong: {}", self.0),
        )
            .into_response()
    }
}

// This enables using `?` on functions that return `Result<_, anyhow::Error>` to turn them into
// `Result<_, AppError>`. That way you don't need to do that manually.
impl<E> From<E> for AppError
where
    E: Into<anyhow::Error>,
{
    fn from(err: E) -> Self {
        Self(err.into())
    }
}

static TEMPLATES_DIR: Dir = include_dir!("$CARGO_MANIFEST_DIR/templates");
static ASSETS_DIR: Dir = include_dir!("$CARGO_MANIFEST_DIR/static");

#[instrument]
fn make_environment() -> Result<Environment<'static>, anyhow::Error> {
    let mut env = Environment::new();
    let mut entries = vec![&TEMPLATES_DIR];

    while let Some(dir) = entries.pop() {
        for entry in dir.entries() {
            match entry {
                DirEntry::Dir(d) => {
                    entries.push(d);
                }
                DirEntry::File(f) => {
                    debug!("adding {:?} to minijinja environment", f.path());
                    env.add_template(
                        f.path().to_str().expect("valid pathname"),
                        f.contents_utf8().expect("contents are valid"),
                    )?;
                }
            }
        }
    }

    info!(
        "loaded {} templates into environment",
        env.templates().count()
    );

    Ok(env)
}

pub fn make_app_router() -> Router {
    Router::new()
        .route("/app/*path", get(app_handler))
        .nest("/api", api::router())
        .route("/assets", get(asset_handler))
        .with_state(AppState::default())
}

#[instrument(skip(ctx))]
async fn app_handler(
    State(ctx): State<AppState>,
    Path(path): Path<String>,
) -> Result<impl IntoResponse, AppError> {
    let template = ctx.templates.get_template(&path.to_string())?;

    let content = template.render("")?;

    let headers = [(header::CONTENT_TYPE, mime::TEXT_HTML_UTF_8.to_string())];

    Ok((StatusCode::OK, headers, content))
}

async fn asset_handler() -> Result<impl IntoResponse, AppError> {
    Ok(StatusCode::OK)
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_environment() -> anyhow::Result<()> {
        make_environment()?;
        Ok(())
    }
}
